package org.gvsig.georeferencing;

import java.awt.geom.Point2D;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.zip.DataFormatException;

import org.cresques.io.RasterMetaFileTags;
import org.gvsig.georeferencing.utils.AffineT;
import org.gvsig.georeferencing.utils.MathUtils;
import org.gvsig.georeferencing.utils.PointT;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;
import org.xmlpull.v1.XmlSerializer;

import com.iver.cit.gvsig.fmap.layers.FLyrPoints;


/** 
 * Operaciones necesarias para la georreferenciacin a partir de la capa de puntos. 
 * En base a unos puntos de control de origen y otros de destino se crea una matriz
 * de transformacin para asignar la nueva posicin a la imagen y crear ficheros 
 * de georreferenciacin en dos formatos: worldfile y rasterMetaFile
 * 
 * @author Nacho Brodin (brodin_ign@gva.es)
 */
public class GeoOperations{
	
	//**********************Vars**********************************
	private int 			order = 1;
	private PointT[] 		src = null;
	private PointT[] 		dst = null;
	private AffineT		affine = null;
	private boolean 	createWorldFile = false;
	//**********************End Vars******************************
	
	//**********************Methods*******************************
	/**
	 * Constructor
	 */
	public GeoOperations(){}
	
	/**
	 * Constructor. Crea las estructuras de puntos y calcula la transformacin afn.
	 * @param lyr
	 */
	public GeoOperations(FLyrPoints lyr){
		FLyrPoints 	lyrPoints = lyr;
		src = new PointT[lyr.getCountActivePoints()]; 
		dst = new PointT[lyr.getCountActivePoints()];
		int nPoint = 0;
		for(int i = 0; i<lyr.getCountActivePoints(); i++){
			if(lyr.getPoint(i).active == true){
				src[nPoint] = new PointT();
				dst[nPoint] = new PointT();
				src[nPoint].setX(lyr.getPoint(i).pixelPoint.getX());
				src[nPoint].setY(lyr.getPoint(i).pixelPoint.getY());
				src[nPoint].setI(lyr.getCountPoints());
				dst[nPoint].setX(lyr.getPoint(i).mapPoint.getX());
				dst[nPoint].setY(lyr.getPoint(i).mapPoint.getY());
				dst[nPoint].setI(lyr.getCountPoints());
				nPoint++;
			}
		}
		
		if(lyr.getCountActivePoints() >= 3 * 10)
			order = 3;
		else if(lyr.getCountActivePoints() >= 3 * 6)
			order = 2;
		else
			order = 1;

		affine = new AffineT();
		
		//Calcular la transformacin afin por mnimos cuadrados
		affine = computeLeastSquaresAffine(affine, src, dst, order);
	}
	
	/**
	 * A partir de la transformacin afn creada en el construtor
	 * genera un fichero de georreferenciacin para la imagen.
	 * @param lyr		Capa de puntos
	 * @param order	Orden del sistema de ecuaciones
	 * @param widthPx	Ancho en pixeles de la imagen a georreferenciar
	 * @param heightPx	Alto en pixeles de la imagen a georreferenciar
	 * @param file		Nombre del fichero raster a georreferenciar
	 */
	public void createGeorefFile(int widthPx, int heightPx, String file){
		try{
			File f = new File(file);
			String nameWorldFile = f.getPath().substring(0, f.getPath().lastIndexOf(".")) + getExtensionWorldFile(file);
			if(createWorldFile)
				createWorldFile(affine, widthPx, heightPx, nameWorldFile);
			createRasterMetaFile(affine, widthPx, heightPx, nameWorldFile.substring(0, nameWorldFile.lastIndexOf("."))+".rmf");
		}catch(IOException ex){
			System.err.println("Can't create WorldFile");
		}
	}
	
	/**
	 * A partir de la transformacin afn calculada en el contructor
	 * transforma los puntos pasados como parmetros.
	 * @param lyr		Capa de puntos
	 * @param list		Lista de puntos a transformar
	 * @return  		Lista de puntos transformados
	 */
	public Point2D[] transformPoints(Point2D[] list){
		Point2D[] result = new Point2D[list.length];
		for(int i = 0; i < list.length;i++){
			double[] pto = transformPoint((int)list[i].getX(), (int)list[i].getY(), affine);
			result[i] = new Point2D.Double(pto[0], pto[1]);
		}
		return result;
	}
	
	/**
	 * Crea un fichero de georreferenciacin (worldfile) a partir de la transformacin afn. Para
	 * esto necesita obtener las coordenadas reales de la coordenada en pixels (0,0) y calcular
	 * el tamao de pixel. 
	 * @param affine		Transformacin
	 * @param widthPx		Ancho en pixeles de la imagen a georreferenciar
	 * @param heightPx		Alto en pixeles de la imagen a georreferenciar
	 * @param nameWordlFile	Nombre del fichero de georreferenciacin
	 * @throws IOException
	 */
	private void createWorldFile(AffineT affine, int widthPx, int heightPx, String nameWordlFile) throws IOException {
		StringBuffer data = new StringBuffer();
		double[] begin = transformPoint(0, 0, affine);
		double[] end = transformPoint(widthPx, heightPx, affine);
		File f = new File(nameWordlFile);
		DataOutputStream dos = new DataOutputStream( new BufferedOutputStream(new FileOutputStream(f)) );
		
		double pixelSizeX = (end[0] - begin[0])/(widthPx - 1);
		double pixelSizeY = (end[1] - begin[1])/(heightPx - 1);
		
    	data.append(pixelSizeX+"\n");
    	data.append("0.0\n");
    	data.append("0.0\n");
    	data.append((-1  * pixelSizeY)+"\n");
    	data.append(""+begin[0]+"\n");
    	data.append(""+end[1]+"\n");
    	
    	dos.writeBytes(data.toString());
		dos.close();
	}
	
	/**
	 * A partir de XmlSerializer se salva la georreferenciacin 
	 * @param serializer
	 * @param min		
	 * @param max		
	 * @param widthPx		Ancho en pixeles de la imagen a georreferenciar
	 * @param heightPx		Alto en pixeles de la imagen a georreferenciar
	 * @throws IOException
	 */
	private void serializeGeoreferencing(XmlSerializer serializer, double[] min, double[] max, int widthPx, int heightPx) throws IOException {
		double pixelSizeX = (max[0] - min[0])/(widthPx - 1);
		double pixelSizeY = (max[1] - min[1])/(heightPx - 1);
		
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PROJ).text("Projection").endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PROJ).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.BBOX).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.POSX).text(""+min[0]).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.POSX).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.POSY).text(""+max[1]).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.POSY).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_SIZE_X).text(""+pixelSizeX).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_SIZE_X).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_SIZE_Y).text(""+ (-1 * pixelSizeY)).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_SIZE_Y).text("\n");		
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.WIDTH).text(""+(max[0] - min[0])).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.WIDTH).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.HEIGHT).text(""+(max[1] - min[1])).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.HEIGHT).text("\n");
		serializer.endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.BBOX).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.DIM).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_WIDTH).text(""+widthPx).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_WIDTH).text("\n");
		serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_HEIGHT).text(""+heightPx).endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.PX_HEIGHT).text("\n");
		serializer.endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.DIM).text("\n");
	}
	
	/**
	 * Si ya existe un fichero de georreferenciacin se ha creado un fichero .tmp con
	 * la nueva georreferenciacin. Esta funcin mezcla este temporal con el .rmf existente
	 * en un nuevo fichero _tmpOutput que ser renombrado como el nuevo .rmf. Finalmente 
	 * elimina el temporal y el _tmpOutput. 
	 * @param fileName
	 */
	private void mergeFiles(String fileName){
		try{
			File rmfFile = new File(fileName);
			File tmpFile = new File(fileName.substring(0, fileName.lastIndexOf("."))+".tmp");
			File out = new File(rmfFile+"_tmpOutput");
			
			BufferedReader inRmf = new BufferedReader(new InputStreamReader(new FileInputStream(rmfFile)));
			BufferedReader inTmp = new BufferedReader(new InputStreamReader(new FileInputStream(tmpFile)));
			BufferedWriter outTmp = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(out)));
			
			//Leemos el principio del .rmf
	        String str = inRmf.readLine();
	        while(!str.startsWith("<"+RasterMetaFileTags.MAIN_TAG)){
	        	outTmp.write(str+"\n");
	        	str = inRmf.readLine();
	        }
	    	outTmp.write(str+"\n");
	        
	        //Leemos la georreferenciacin del .tmp
	        while(str != null && !str.startsWith("<"+RasterMetaFileTags.LAYER))
	        	str = inTmp.readLine();
	        while(str != null){
	        	outTmp.write(str+"\n");
	        	str = inTmp.readLine();
	        }
	        	
	        //Saltamos la georreferenciacin que ya existia en el .rmf
	        try{
	        	str = inRmf.readLine();
		        while(str != null && 
		        	  !str.startsWith("</"+RasterMetaFileTags.LAYER) &&
		        	  !str.startsWith("<"+RasterMetaFileTags.GEOPOINTS) &&
		        	  !str.startsWith("</"+RasterMetaFileTags.MAIN_TAG))
		        	str = inRmf.readLine();
	        }catch(Exception exc){
	        	
	        }
	        
	        //Leemos el resto del fichero .rmf
	        if(!str.startsWith("<"+RasterMetaFileTags.GEOPOINTS))
	        	str = inRmf.readLine();
	        while(str != null){
	        	outTmp.write(str+"\n");
	        	str = inRmf.readLine();
	        }
	        
	        outTmp.close();
	        inRmf.close();
	        inTmp.close();
	        
	        //Eliminamos el antiguo .rmf y lo sustituimos
	        rmfFile.delete();
	        out.renameTo(rmfFile);
	        tmpFile.delete();
	        
		}catch(FileNotFoundException exc){
			
		}catch(IOException exc){
			
		}
	}
	
	/**
	 * Si no existe crea un fichero .rmf con la georreferenciacin de la imagen. Si existe este 
	 * deber aadir o modificar la informacin de georreferenciacin en el fichero.
	 * Para esto necesita obtener las coordenadas reales de la coordenada en pixels (0,0) y 
	 * calcular el tamao de pixel. 
	 * @param affine		Transformacin
	 * @param widthPx		Ancho en pixeles de la imagen a georreferenciar
	 * @param heightPx		Alto en pixeles de la imagen a georreferenciar
	 * @param nameWordlFile	Nombre del fichero de georreferenciacin
	 * @throws IOException
	 */
	private void createRasterMetaFile(AffineT affine, int widthPx, int heightPx, String nameWorldFile) throws IOException {
		File file = new File(nameWorldFile);
		double[] min = transformPoint(0, 0, affine);
		double[] max = transformPoint(widthPx, heightPx, affine);
		try{		
			XmlPullParserFactory factory = XmlPullParserFactory.newInstance(System.getProperty(XmlPullParserFactory.PROPERTY_NAME), null);
			XmlSerializer serializer = factory.newSerializer();
			
			if(file.exists()){
				file = new File(nameWorldFile.substring(0, nameWorldFile.lastIndexOf("."))+".tmp");
				serializer.setOutput(new FileOutputStream(file), null);
				serializer.startDocument(null, null);
				serializer.ignorableWhitespace("\n\n");
				serializer.setPrefix("", RasterMetaFileTags.NAMESPACE);
				serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.LAYER).text("\n");
				serializeGeoreferencing(serializer, min, max, widthPx, heightPx);
				serializer.endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.LAYER).text("\n");
				serializer.endDocument();
				mergeFiles(nameWorldFile);
			}else{
					serializer.setOutput(new FileOutputStream(file), null);
					serializer.startDocument(null, null);
					serializer.ignorableWhitespace("\n\n");
					serializer.setPrefix("", RasterMetaFileTags.NAMESPACE);
					serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.MAIN_TAG).text("\n");
					serializer.startTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.LAYER).text("\n");
					serializeGeoreferencing(serializer, min, max, widthPx, heightPx);
					serializer.endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.LAYER).text("\n");
					serializer.endTag(RasterMetaFileTags.NAMESPACE, RasterMetaFileTags.MAIN_TAG).text("\n");
					serializer.endDocument();
			}
		}catch(XmlPullParserException exc){}
	}
		
	/**
	 * Calcula la transformacin afn a partir de los puntos introducidos por el mtodo de 
	 * mnimos cuadrados.
	 * @param af	Transformacin afn
	 * @param src	Puntos de entrada en coordenadas pixel
	 * @param dst	Puntos de destino en coordenadas del mundo 	
	 * @param order	Orden del sistema de ecuaciones
	 */
	public AffineT computeLeastSquaresAffine(AffineT af, PointT[] src, PointT[] dst, int order){
		double[] b = new double[dst.length];
		int i,cofs;

		af.setOrder(order);
		cofs = order <= 2 ? order*3 : 10;
		af.mallocCofs(cofs);
		
		//First compute the X cofs  
		for(i = 0; i < dst.length; i++) {    
			b[i] = dst[i].getX();  
		}
		
		af.setCofX(singleLeastSquaresAffine(af.getXcofs(), src, b, order));
		
		//Now compute the Y cofs  
		for(i = 0; i < dst.length; i++) {    
			b[i] = dst[i].getY();  
		}
		
		af.setCofY(singleLeastSquaresAffine(af.getYcofs(), src, b, order));
		
		return af;
	}
	
	private double[] singleLeastSquaresAffine(double[] a, PointT[] src, double[] dst, int order){
		int i,j;
		int n = dst.length;

		int points = order <= 2 ? order * 3 : 10;	
		double[][] combined = new double[points][points + 1];
		double[][] A = new double[n + 1][points];
		double[][] b = new double[n + 1][1];
		
		for(i = 0; i < n; i++) {
			A[i][0] = 1.0;
			A[i][1] = src[i].getX();
			A[i][2] = src[i].getY();
			if(order > 1) {
				A[i][3] = src[i].getX() * src[i].getX();
				A[i][4] = src[i].getX() * src[i].getY();
				A[i][5] = src[i].getY() * src[i].getY();
			}
			if(order > 2) {
				A[i][6] = src[i].getX() * src[i].getX() * src[i].getX();
				A[i][7] = src[i].getX() * src[i].getX() * src[i].getY();
				A[i][8] = src[i].getX() * src[i].getY() * src[i].getY();
				A[i][9] = src[i].getY() * src[i].getY() * src[i].getY();
			}
			b[i][0] = dst[i];
		}
		
		double[][] Atrans = MathUtils.transpose(A, n, points);
		double[][] left = MathUtils.multmatrix(Atrans,A,points,n,points);
		double[][] right = MathUtils.multmatrix(Atrans,b,points,n,1);
		
		for(i = 0; i < points; i++) {
			combined[i][0] = right[i][0];
			for(j = 0; j < points; j++) {
				combined[i][j+1] = left[i][j];
			}
		}
		
		try{
			combined = solve(combined, points, points+1);
		}catch(DataFormatException ex){
			System.err.println("Can't solve matrix");
		}

		for(i = 0; i < points; i++) {
			a[i] = combined[i][0];
		}
		return a;
	}
	
	
	private double[][] solve(double[][] mat,int rows,int cols)throws DataFormatException{
		int i,j,k;
		double d,tmp;
		int big;

		for(i = 0; i < rows; i++) {
			// Find largest row
				big = i;
				for(j = i; j < rows; j++) {
					if(Math.abs(mat[j][i+1]) > Math.abs(mat[big][i+1])) {
						big = j;
					}
				}
				// swap row i and row big
				for(k = 0; k < cols ; k++) {
					tmp = mat[i][k];
					mat[i][k] = mat[big][k];
					mat[big][k] = tmp;
				}
			if(mat[i][i+1] == 0) 
				throw new DataFormatException();
			
			d = 1.0 / mat[i][i+1]; 
			for(j = 0; j < cols ; j++) {
				mat[i][j] *= d;
				//assert(!isnan(mat[i][j]));
			}
			for(k = 0; k < rows; k++) {
				if(k == i)
					continue;
				if(mat[k][i+1] != 0.0) {
				d = mat[k][i+1] / mat[i][i+1]; 
					for(j = 0; j < cols ; j++) {
						mat[k][j] -= d*mat[i][j];
						//assert(!isnan(mat[k][j]));
					}
				}
			}
		}
		return mat;
	}
	
	/**
	 * Obtiene las coordenadas de un punto en coordenadas del mundo a partir de una transformacin 
	 * y el punto de entrada en coordenadas de la imagen.
	 * @param sx	Coordenada X del punto de entrada 
	 * @param syz	Coordenada Y del punto de entrada
	 * @param af	Transformacin
	 * @return		Punto transformado
	 */
	public double[] transformPoint(int sx, int sy, AffineT af){
		double[] p = new double[2];
		if(af.getOrder() == 1) {
			p[0] = af.getCofX(0) + af.getCofX(1) * sx + af.getCofX(2) * sy;
			p[1] = af.getCofY(0) + af.getCofY(1) * sx + af.getCofY(2) * sy;
		}
		else if(af.getOrder() == 2) {
			p = quadTransformPoint(sx, sy, af);
		}
		else {
			p = cubicTransformPoint(sx, sy, af);
	 	}
		return p;
	}
	
	private static double[] quadTransformPoint(int sx, int sy, AffineT af){
		double[] p = new double[2];
		p[0] = af.getCofX(0) + af.getCofX(1) * sx + af.getCofX(2) * sy + 
			af.getCofX(3) * sx * sx + af.getCofX(4) * sx * sy + af.getCofX(5) * sy * sy;
		p[1] = af.getCofY(0) + af.getCofY(1) * sx + af.getCofY(2) * sy + 
			af.getCofY(3) * sx * sx + af.getCofY(4) * sx * sy + af.getCofY(5) * sy * sy;
		return p;
	}
	
	private static double[] cubicTransformPoint(int sx, int sy, AffineT af){
		double[] p = new double[2];
		p[0] = af.getCofX(0) + af.getCofX(1) * sx + af.getCofX(2) * sy + 
			af.getCofX(3) * sx * sx + af.getCofX(4) * sx * sy + af.getCofX(5) * sy * sy +
			af.getCofX(6) * sx * sx * sx + af.getCofX(7) * sx * sx * sy +
			af.getCofX(8)*sx*sy*sy + af.getCofX(9)*sy*sy*sy;
		p[1] = af.getCofY(0) + af.getCofY(1) * sx + af.getCofY(2) * sy + 
			af.getCofY(3) * sx * sx + af.getCofY(4) * sx * sy + af.getCofY(5) * sy * sy +
			af.getCofY(6) * sx * sx * sx + af.getCofY(7) * sx * sx * sy +
			af.getCofY(8) * sx * sy * sy + af.getCofY(9) * sy * sy * sy;
		return p;
	}	
	//**********************End Methods***************************
	
	//**********************Setters & Getters*********************
	/**
	 * Mtodo para notificar si se desea crear o no el fichero de coordenadas .tfw, .wld .jpgw ,... 
	 * @param createWorldFile
	 */
	public void setCreateWorldFile(boolean createWorldFile) {
		this.createWorldFile = createWorldFile;
	}
	
	/**
	 * Obtiene la extensin del fichero de georreferenciacin
	 * @return String con la extensi�n del fichero de georreferenciaci�n dependiendo
	 * del valor del formato obtenido del servidor. Por defecto asignaremos un .wld 
	 */
	private String getExtensionWorldFile(String file){
		String ext = file.substring(file.lastIndexOf(".") + 1);
		String extWorldFile = ".wld";
    	if(ext.equals("tif") || ext.equals("tiff"))
    		extWorldFile = ".tfw";
    	if(ext.equals("jpeg") || ext.equals("jpg"))
    		extWorldFile = ".jpgw";
    	return extWorldFile;
	}
	
	/**
	 * Obtiene la matriz de transformacin
	 * @return AffineT
	 */
	public AffineT getAffine() {
		return affine;
	}
	//**********************End Setters & Getters*****************
	
}