/*
 * Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 *
 * Copyright (C) 2004-5.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 * cresques@gmail.com
 */
package org.cresques.io;

import java.awt.Dimension;
import java.awt.Image;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;

import org.cresques.cts.ICoordTrans;
import org.cresques.cts.IProjection;
import org.cresques.px.Extent;

import com.ermapper.ecw.JNCSFile;
import com.ermapper.ecw.JNCSFileNotOpenException;
import com.ermapper.ecw.JNCSInvalidSetViewException;
import com.ermapper.ecw.JNCSProgressiveUpdate;
import com.ermapper.util.JNCSDatasetPoint;
import com.ermapper.util.JNCSWorldPoint;


/**
 * Soporte para los ficheros .ecw de ErMapper.
 * <br>
 * NOTA: El SDK que ermapper ha puesto a disposicin del pblico en java
 * es una versin 2.45, de 19/11/2001. Est implementada usando JNI que
 * se apoya en tres libreras dinmicas (dll), y presenta deficiencias
 * muy graves a la hora de acceder a la informacin. Hasta el momento
 * hemos detectado 3 de ellas:<BR>
 *     1.- No soporta ampliaciones superiores a 1:1. si se intenta acceder
 * a un ecw con un zoom mayor da una excepcin del tipo
 * com.ermapper.ecw.JNCSInvalidSetViewException, que de no ser tenida encuenta
 * acaba tirando abajo la mquina virtual de java.<BR>
 *     2.- La longitud mxima de lnea que adminte el mtodo readLineRGBA es
 * de unos 2500 pixeles, lo que hace el uso para la impresin en formatos
 * superiorea a A4 a 300 ppp o ms inviable.<BR>
 *     3.- La actualizacin progresiva usando el interface JNCSProgressiveUpdate
 * con el JNCSFile hace que el equipo genere un error severo y se apague. Este
 * error imposibilita esta tcnica de acceso a ECW.<BR>
 * <br>
 * Para saltarnos la limitacin del bug#1 pedimos la ventana correspondiente al zoom 1:1 para
 * el view que nos han puesto, y la resizeamos al tamao que nos pide el usuario.<br>
 * Como consecuencia del bug#2, para tamaos de ventana muy grandes (los necesarios
 * para imprimir a ms de A4 a 300DPI), hay que hacer varias llamadas al fichero con
 * varios marcos contiguos, y los devolvemos 'pegados' en una sola imagen (esto se
 * realiza de manera transparente para el usuario dentro de la llamada a updateImage.<br>
 *
 * @author "Luis W. Sevilla" <sevilla_lui@gva.es>
 */
public class EcwFile extends GeoRasterFile implements JNCSProgressiveUpdate {
    //Lleva la cuenta del nmero de actualizaciones que se hace de una imagen que corresponde con el 
    //nmero de bandas que tiene. Esto es necesario ya que si una imagen tiene el mustResize a
    //true solo debe llamar a la funcin resizeImage al actualizar la ltima banda, sino al hacer
    //un zoom menor que 1:1 se veria mal
    private static int nUpdate = 0;
    private JNCSFile file = null;
    private boolean bErrorOnOpen = false;
    private String errorMessage = null;
    private boolean multifile = false;
    private Extent v = null;

    // Ultimo porcentaje de refresco. Se carga en el update y se
    // actualiza en el refreshUpdate
    private int lastRefreshPercent = 0;

    public EcwFile(IProjection proj, String fName) {
        super(proj, null);
        fName = DataSource.normalize(fName);
        super.setName(fName);
        extent = new Extent();

        try {
            //System.err.println("Abriendo "+fName);
            file = new JNCSFile(fName, false);
            load();
            //readGeoInfo(fName);
            bandCount = file.numBands;

            if ( bandCount > 2) {
                    setBand(RED_BAND,   0);
                    setBand(GREEN_BAND, 1);
                    setBand(BLUE_BAND,  2);
            } else
                    setBand(RED_BAND|GREEN_BAND|BLUE_BAND, 0);
        } catch (Exception e) {
            bErrorOnOpen = true;
            errorMessage = e.getMessage();
            System.err.println(errorMessage);
            e.printStackTrace();
        }
    }

    /**
     * Carga un ECW.
     *
     * @param fname
     */
    public GeoFile load() {
        double minX;
        double minY;
        double maxX;
        double maxY;

        if(file.cellIncrementY > 0)
        	file.cellIncrementY = -file.cellIncrementY;
        
        minX = file.originX;
        maxY = file.originY;
        maxX = file.originX +
               ((double) (file.width - 1) * file.cellIncrementX);
        minY = file.originY +
               ((double) (file.height - 1) * file.cellIncrementY);
        
        extent = new Extent(minX, minY, maxX, maxY);
        return this;
    }

    public void close() {
    	if(file != null){
    		file.close(true);
    		file = null;
    	}
    }

    /**
     * Devuelve el ancho de la imagen
     */
    public int getWidth() {
        return file.width;
    }

    /**
     * Devuelve el alto de la imagen
     */
    public int getHeight() {
        return file.height;
    }

    /**
     *
     */
    public void setMultifile(boolean mult) {
        this.multifile = mult;
    }

    public void setView(Extent e) {
    	//Aplicamos la transformacin a la vista en caso de que haya un fichero .rmf 
    	/*    	         
    	if(file.cellIncrementY > 0)
        	file.cellIncrementY = -file.cellIncrementY;
    	if(minX < file.originX)
    		minX = file.originX;
    	if(maxY > file.originY)
    		maxY = file.originY;
    	if(maxX > (file.originX + ((double) (file.width - 1) * file.cellIncrementX)))
    		maxX = file.originX + ((double) (file.width - 1) * file.cellIncrementX);
    	if(minY < file.originY + ((double) (file.height - 1) * file.cellIncrementY))
    		minY = file.originY + ((double) (file.height - 1) * file.cellIncrementY);
    	
		Extent transformView = new Extent(	minX, minY, maxX, maxY );*/
        v = new Extent(e);
    }

    public Extent getView() {
        return v;
    }
    
    private void setFileView(int numBands, int [] bandList, ChunkFrame f)
    	throws JNCSFileNotOpenException, JNCSInvalidSetViewException {
        file.setView(file.numBands, bandList, f.v.minX(), f.v.maxY(), f.v.maxX(), f.v.minY(), f.width, f.height);
    }

    /**
     * Obtiene un trozo de imagen (determinado por la vista y los parmetros.
     *
     * @param width
     * @param height
     */
    public synchronized Image updateImage(int width, int height, ICoordTrans rp) {
        // TODO reproyectar para devolver el trozo de imagen pedida sobre ...
        // la proyeccin de destino.
        int line = 0;
        boolean mustResize = false;
        Dimension fullSize = null;
        Image ecwImage = null;

        if (file == null) {
            return ecwImage;
        }

        try {
            int[] bandlist;
            int[] bandListTriband;
            int[] pRGBArray = null;

            if(mustVerifySize()){
            // Work out the correct aspect for the setView call.
	            double dFileAspect = (double) v.width() / (double) v.height();
	            double dWindowAspect = (double) width / (double) height;
	
	            if (dFileAspect > dWindowAspect) {
	                height = (int) ((double) width / dFileAspect);
	            } else {
	                width = (int) ((double) height * dFileAspect);
	            }
            }
            fullSize = new Dimension(width, height);

            //System.out.println("fullSize = ("+width+","+height+")");
            // Peta en los pequeos ... arreglar antes de meter del todo
            ChunkFrame[] frames = ChunkFrame.computeFrames(file, v, fullSize, extent);

            if (frames.length == 1) {
                width = frames[0].width;
                height = frames[0].height;

                if (width <= 0) {
                    width = 1;
                }

                if (height <= 0) {
                    height = 1;
                }
            }

            /*                        JNCSDatasetPoint ptMin = file.convertWorldToDataset(v.minX(), v.minY());
                                    JNCSDatasetPoint ptMax = file.convertWorldToDataset(v.maxX(), v.maxY());
                                    System.out.println("Dataset coords Width = "+(ptMax.x-ptMin.x)+", px width ="+width);
                                    // BEGIN Cambiando para soportar e < 1:1
                                    // TODO Mejorarlo para que los PAN con un zoom muy grande sean correctos
                                    if ((ptMax.x-ptMin.x)<width) {
                                            width = ptMax.x-ptMin.x;
                                            height = ptMin.y-ptMax.y;
                                            System.out.println("Size=("+width+","+height+")");
                                            mustResize = true;
                                    }*/

            // Create an image of the ecw file.
            if (doTransparency) {
                ecwImage = new BufferedImage(width, height,
                                             BufferedImage.TYPE_INT_ARGB);
            } else {
                ecwImage = new BufferedImage(width, height,
                                             BufferedImage.TYPE_INT_RGB);
            }

            pRGBArray = new int[width];

            // Setup the view parameters for the ecw file.
            bandlist = new int[bandCount];
            bandListTriband = new int[bandCount];

            if (bandCount > 2) {
                bandlist[0] = getBand(RED_BAND);
                bandlist[1] = getBand(GREEN_BAND);
                bandlist[2] = getBand(BLUE_BAND);

                if (bandCount > 3) {
                    for (int i = 3; i < bandCount; i++) {
                        bandlist[i] = 0;
                    }
                }
            } else {
                for (int i = 0; i < bandCount; i++) {
                    bandlist[i] = i;
                }
            }

            if (bandCount == 3) {
                bandListTriband[0] = 0;
                bandListTriband[1] = 1;
                bandListTriband[2] = 2;
            }

            for (int nChunk = 0; nChunk < frames.length; nChunk++) {
                ChunkFrame f = frames[nChunk];

                // Set the view			
                if (bandCount != 3) {
                    setFileView(file.numBands, bandlist, f);
                } else {
                    setFileView(file.numBands, bandListTriband, f);
                }

                /* 
                 * Esta peli es porque el Ecw no intercambia las bandas con lo que me toca hacerlo
                 * a mano. Primero detectamos si se ha alterado el orden de las mismas. Si es as
                 * calculamos mascaras y desplazamientos y hacemos una copia en pRGBArrayCopy
                 * con las bandas alteradas de orden
                 */
                int[] pRGBArrayCopy = null;
                int[] mascara = new int[3];
                int[] shl = new int[3];
                int[] shr = new int[3];
                boolean order = true;

                if (bandCount == 3) {
                    for (int i = 0; i < bandCount; i++)
                        if (bandlist[i] != i) {
                            order = false;
                        }

                    if (!order) {
                        for (int i = 0; i < bandCount; i++) {
                            switch (bandlist[i]) {
                            case 0:
                                mascara[i] = 0x00ff0000;
                                break;
                            case 1:
                                mascara[i] = 0x0000ff00;
                                break;
                            case 2:
                                mascara[i] = 0x000000ff;
                                break;
                            }
                            if ((i == 1) && (bandlist[i] == 0)) 
                                shr[i] = 8;
                            if ((i == 2) && (bandlist[i] == 0)) 
                                shr[i] = 16;
                            if ((i == 0) && (bandlist[i] == 1)) 
                                shl[i] = 8;
                            if ((i == 2) && (bandlist[i] == 1))
                                shr[i] = 8;
                            if ((i == 0) && (bandlist[i] == 2)) 
                                shl[i] = 16;
                            if ((i == 1) && (bandlist[i] == 2))
                                shl[i] = 8;
                        }
                    }
                }

                // Read the scan lines
                for (line = 0; line < f.height; line++) {
                    file.readLineRGBA(pRGBArray);

                    if ((bandCount == 3) && !order) {
                        pRGBArrayCopy = new int[pRGBArray.length];

                        for (int i = 0; i < pRGBArray.length; i++) {
                            pRGBArrayCopy[i] = (pRGBArray[i] & 0xff000000) +
                                               (((pRGBArray[i] & mascara[2]) << shl[2]) >> shr[2]) +
                                               (((pRGBArray[i] & mascara[1]) << shl[1]) >> shr[1]) +
                                               (((pRGBArray[i] & mascara[0]) << shl[0]) >> shr[0]);
                        }

                        pRGBArray = pRGBArrayCopy;
                    }

                    // Prueba de sustitucin de color transparente
                    if (doTransparency) {
                        if (line == 0) {
                            tFilter.debug = true;
                        }

                        tFilter.filterLine(pRGBArray);
                        tFilter.debug = false;
                    }

                    ((BufferedImage) ecwImage).setRGB(f.pos.x, f.pos.y + line,
                                                      f.width, 1, pRGBArray, 0,
                                                      f.width);
                }
            }

            if (frames[0].mustResize) {
                //System.out.println("resize "+fullSize);
                return resizeImage(fullSize, ecwImage);
            }

            /*
             * La excepcin atrapada es la de 'zoom > 1:1 no valido'
            } catch (com.ermapper.ecw.JNCSInvalidSetViewException e) {
                    System.err.println(errorMessage);
                    e.printStackTrace(); */
        } catch (com.ermapper.ecw.JNCSException e) { //java.lang.ArrayIndexOutOfBoundsException:
            bErrorOnOpen = true;
            System.err.println("EcwFile JNCS Error en la lnea " + line + "/" +
                               height);
            System.err.println(e.getMessage());
            e.printStackTrace();
        } catch (java.lang.ArrayIndexOutOfBoundsException e) { //:
            bErrorOnOpen = true;
            System.err.println("EcwFile ArrayIndex Error en la lnea " + line +
                               "/" + height);
            System.err.println(e.getMessage());
            e.printStackTrace();
        } catch (Exception e) {
            bErrorOnOpen = true;
            errorMessage = e.getMessage();

            //			g.drawString(errorMessage, 0, 50);
            System.err.println(errorMessage);
            e.printStackTrace();
        }

        lastRefreshPercent = file.getPercentComplete();
        System.out.println("Leido al " + lastRefreshPercent + " %.");

        return ecwImage;
    }

    /**
     * Redimensionado de imagen
     * La funcin getScaledInstance nos devuelve un tipo image que no sirve por lo que
     * habr que crear  buffImg como BufferedImage y copiar los datos devueltos por esta
     * funcin a este que es el que ser devuelto por la funcin
     * @param sz
     * @param image        Image de entrada
     * @return        Imagen reescalada
     */
    private Image resizeImage(Dimension sz, Image image) {
        Image buffImg = null;
        Image img = image.getScaledInstance((int) sz.getWidth(),
                                            (int) sz.getHeight(),
                                            Image.SCALE_SMOOTH);

        //Todo este pollo es para copiar el tipo image devuelto a BufferedImage
        buffImg = new BufferedImage(img.getWidth(null), img.getHeight(null),
                                    BufferedImage.TYPE_INT_ARGB);

        int[] pixels = new int[img.getWidth(null) * img.getHeight(null)];
        PixelGrabber pg = new PixelGrabber(img, 0, 0, img.getWidth(null),
                                           img.getHeight(null), pixels, 0,
                                           img.getWidth(null));

        try {
            pg.grabPixels();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        for (int j = 0; j < buffImg.getHeight(null); j++) {
            for (int i = 0; i < buffImg.getWidth(null); i++) {
                ((BufferedImage) buffImg).setRGB(i, j,
                                                 pixels[(j * buffImg.getWidth(null)) +
                                                 i]);
            }
        }

        return buffImg;
    }

    /**
     * Reproyecta el raster.
     */
    public void reProject(ICoordTrans rp) {
        // TODO metodo reProject pendiente de implementar
    }

    /**
     * Soporte para actualizacin de la imagen
     * @see com.ermapper.ecw.JNCSProgressiveUpdate#refreshUpdate(int, int, double, double, double, double)
     */
    public void refreshUpdate(int nWidth, int nHeight, double dWorldTLX,
                              double dWorldTLY, double dWorldBRX,
                              double dWorldBRY) {
        int completado = file.getPercentComplete();
        System.out.println("EcwFile: se actualiza 1: " + completado +
                           " % completado");

        if ((updatable != null) && (lastRefreshPercent < 100)) {
            if (((completado - lastRefreshPercent) > 25) ||
                    (completado == 100)) {
                lastRefreshPercent = file.getPercentComplete();
                updatable.repaint();
            }
        }
    }

    public void refreshUpdate(int nWidth, int nHeight, int dDatasetTLX,
                              int dDatasetTLY, int dDatasetBRX, int dDatasetBRY) {
        System.out.println("EcwFile: se actualiza 2");
    }

    /**
     *  Esta funcin es porque el Ecw no intercambia las bandas con lo que me toca hacerlo
     * a mano. Primero detectamos si se ha alterado el orden de las mismas. Si es as
     * calculamos mascaras y desplazamientos y hacemos una copia en pRGBArrayCopy
     * con las bandas alteradas de orden
     * @param bandList        lista de bandas
     * @param mask mascara
     * @param shl desplazamiento izquierda
     * @param shr desplazamiento derecha
     */
    private boolean calcMaskAndShift(int[] bandList, int[] mask, int[] shl,
                                     int[] shr) {
        boolean order = true;

        if (bandCount == 3) {
            for (int i = 0; i < bandCount; i++)
                if (bandList[i] != i) {
                    order = false;
                }

            if (!order) {
                for (int i = 0; i < bandCount; i++) {
                    switch (bandList[i]) {
                    case 0:
                        mask[i] = 0x00ff0000;
                        break;
                    case 1:
                        mask[i] = 0x0000ff00;
                        break;
                    case 2:
                        mask[i] = 0x000000ff;
                        break;
                    }
                    
                    if ((i == 1) && (bandList[i] == 0))
                        shr[i] = 8;
                    if ((i == 2) && (bandList[i] == 0)) 
                        shr[i] = 16;
                    if ((i == 0) && (bandList[i] == 1))
                        shl[i] = 8;
                    if ((i == 2) && (bandList[i] == 1))
                        shr[i] = 8;
                    if ((i == 0) && (bandList[i] == 2))
                        shl[i] = 16;
                    if ((i == 1) && (bandList[i] == 2))
                        shl[i] = 8;
                }
            }
        }

        return order;
    }

    /**
     * Intercambio de bandas para ecw manual. Se le pasa el array de bytes que se desea intercambiar
     * la mascara y desplazamientos previamente calculados con calcMaskAndShift
     * @param order        true si el orden de las bandas no est alterado y false si lo est
     * @param pRGBArray array de
     * @param mascara
     * @param shl desplazamiento a izquierda
     * @param shr desplazamiento a derecha
     * @return array con las bandas cambiadas
     */
    private int[] changeBands(boolean order, int[] pRGBArray, int[] mascara,
                              int[] shl, int[] shr) {
        if ((bandCount == 3) && !order) {
            int[] pRGBArrayCopy = new int[pRGBArray.length];

            for (int i = 0; i < pRGBArray.length; i++) {
                pRGBArrayCopy[i] = (pRGBArray[i] & 0xff000000) +
                                   (((pRGBArray[i] & mascara[2]) << shl[2]) >> shr[2]) +
                                   (((pRGBArray[i] & mascara[1]) << shl[1]) >> shr[1]) +
                                   (((pRGBArray[i] & mascara[0]) << shl[0]) >> shr[0]);
            }

            return pRGBArrayCopy;
        }

        return pRGBArray;
    }

    /**
     * Asigna al objeto Image los valores con los dato de la imagen contenidos en el
     * vector de enteros.
     * @param image        imagen con los datos actuales
     * @param startX        inicio de la posicin en X dentro de la imagen
     * @param startY        inicio de la posicin en X dentro de la imagen
     * @param w        Ancho de la imagen
     * @param h        Alto de la imagen
     * @param rgbArray        vector que contiene la banda que se va a sustituir
     * @param offset        desplazamiento
     * @param scansize        tamao de imagen recorrida por cada p
     */
    protected void setRGBLine(BufferedImage image, int startX, int startY,
                              int w, int h, int[] rgbArray, int offset,
                              int scansize) {
        image.setRGB(startX, startY, w, h, rgbArray, offset, scansize);
    }

    /**
     * Asigna al objeto Image la mezcla entre los valores que ya tiene y los valores
     * con los dato de la imagen contenidos en el vector de enteros. De los valores RGB
     * que ya contiene se mantienen las bandas que no coinciden con el valor de flags. La
     * banda correspondiente a flags es sustituida por los datos del vector.
     * @param image        imagen con los datos actuales
     * @param startX        inicio de la posicin en X dentro de la imagen
     * @param startY        inicio de la posicin en X dentro de la imagen
     * @param w        Ancho de la imagen
     * @param h        Alto de la imagen
     * @param rgbArray        vector que contiene la banda que se va a sustituir
     * @param offset        desplazamiento
     * @param scansize        tamao de imagen recorrida por cada paso
     * @param flags        banda que se va a sustituir (Ctes de GeoRasterFile)
     */
    protected void setRGBLine(BufferedImage image, int startX, int startY,
                              int w, int h, int[] rgbArray, int offset,
                              int scansize, int flags) {
        int[] line = new int[rgbArray.length];
        image.getRGB(startX, startY, w, h, line, offset, scansize);

        if (flags == GeoRasterFile.RED_BAND) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0x0000ffff) | (rgbArray[i] & 0xffff0000);
        } else if (flags == GeoRasterFile.GREEN_BAND) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0x00ff00ff) | (rgbArray[i] & 0xff00ff00);
        } else if (flags == GeoRasterFile.BLUE_BAND) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0x00ffff00) | (rgbArray[i] & 0xff0000ff);
        }

        image.setRGB(startX, startY, w, h, line, offset, scansize);
    }

    /**
     * Asigna al objeto Image la mezcla entre los valores que ya tiene y los valores
     * con los dato de la imagen contenidos en el vector de enteros. De los valores RGB
     * que ya contiene se mantienen las bandas que no coinciden con el valor de flags. La
     * banda correspondiente a flags es sustituida por los datos del vector.
     * @param image        imagen con los datos actuales
     * @param startX        inicio de la posicin en X dentro de la imagen
     * @param startY        inicio de la posicin en X dentro de la imagen
     * @param w        Ancho de la imagen
     * @param h        Alto de la imagen
     * @param rgbArray        vector que contiene la banda que se va a sustituir
     * @param offset        desplazamiento
     * @param scansize        tamao de imagen recorrida por cada paso
     * @param origBand        Banda origen del GeoRasterFile
     * @param destBandFlag        banda que se va a sustituir (Ctes de GeoRasterFile)
     */
    protected void setRGBLine(BufferedImage image, int startX, int startY,
                              int w, int h, int[] rgbArray, int offset,
                              int scansize, int origBand, int destBandFlag) {
        int[] line = new int[rgbArray.length];
        image.getRGB(startX, startY, w, h, line, offset, scansize);

        if ((origBand == 0) && (destBandFlag == GeoRasterFile.RED_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xff00ffff) | (rgbArray[i] & 0x00ff0000);
        } else if ((origBand == 1) &&
                       (destBandFlag == GeoRasterFile.GREEN_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xffff00ff) | (rgbArray[i] & 0x0000ff00);
        } else if ((origBand == 2) &&
                       (destBandFlag == GeoRasterFile.BLUE_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xffffff00) | (rgbArray[i] & 0x000000ff);
        }
        else if ((origBand == 0) && (destBandFlag == GeoRasterFile.GREEN_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xffff00ff) |
                          ((rgbArray[i] & 0x00ff0000) >> 8);
        } else if ((origBand == 0) &&
                       (destBandFlag == GeoRasterFile.BLUE_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xffffff00) |
                          ((rgbArray[i] & 0x00ff0000) >> 16);
        }
        else if ((origBand == 1) && (destBandFlag == GeoRasterFile.RED_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xff00ffff) |
                          ((rgbArray[i] & 0x0000ff00) << 8);
        } else if ((origBand == 1) &&
                       (destBandFlag == GeoRasterFile.BLUE_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xffffff00) |
                          ((rgbArray[i] & 0x0000ff00) >> 8);
        }
        else if ((origBand == 2) && (destBandFlag == GeoRasterFile.RED_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xff00ffff) |
                          ((rgbArray[i] & 0x000000ff) << 16);
        } else if ((origBand == 2) &&
                       (destBandFlag == GeoRasterFile.GREEN_BAND)) {
            for (int i = 0; i < line.length; i++)
                line[i] = (line[i] & 0xffff00ff) |
                          ((rgbArray[i] & 0x000000ff) << 8);
        }

        image.setRGB(startX, startY, w, h, line, offset, scansize);
    }
    
    
    /* (non-Javadoc)
     * @see org.cresques.io.GeoRasterFile#updateImage(int, int, org.cresques.cts.ICoordTrans, java.awt.Image, int origBand, int destBand)
     */
    public Image updateImage(int width, int height, ICoordTrans rp, Image img,
                             int origBand, int destBandFlag) throws SupersamplingNotSupportedException{
        //TODO reproyectar para devolver el trozo de imagen pedida sobre ...
        // la proyeccin de destino.
        int line = 0;
        boolean mustResize = false;
        Dimension fullSize = null;
        boolean trySupersampling = false;

        if (file == null) {
            return null;
        }

        try {
            int[] bandlist;
            int[] bandListTriband;
            int[] pRGBArray = null;

            if(mustVerifySize()){
	            // Work out the correct aspect for the setView call.
	            double dFileAspect = (double) v.width() / (double) v.height();
	            double dWindowAspect = (double) width / (double) height;
	
	            if (dFileAspect > dWindowAspect) {
	                height = (int) ((double) width / dFileAspect);
	            } else {
	                width = (int) ((double) height * dFileAspect);
	            }
            }

            fullSize = new Dimension(width, height);
            
            ChunkFrame[] frames = ChunkFrame.computeFrames(file, v, fullSize, extent);
            	
            if (frames.length == 1) {
                width = frames[0].width;
                height = frames[0].height;

                if (width <= 0)
                    width = 1;
             
                if (height <= 0)
                    height = 1;
            }
                        
            // Create an image of the ecw file.
            pRGBArray = new int[width];

            // Setup the view parameters for the ecw file.
            bandlist = new int[bandCount];
            bandListTriband = new int[bandCount];

            if (bandCount > 2) {
                bandlist[0] = getBand(RED_BAND);
                bandlist[1] = getBand(GREEN_BAND);
                bandlist[2] = getBand(BLUE_BAND);

                if (bandCount > 3) {
                    for (int i = 3; i < bandCount; i++) {
                        bandlist[i] = 0;
                    }
                }
            } else {
                for (int i = 0; i < bandCount; i++)
                    bandlist[i] = i;
            }

            if (bandCount == 3) {
                bandListTriband[0] = 0;
                bandListTriband[1] = 1;
                bandListTriband[2] = 2;
            }

            int[] mascara = new int[3];
            int[] shl = new int[3];
            int[] shr = new int[3];
            boolean order = true;
            
            if (img == null) { //Caso en el que se crea un Image
                EcwFile.nUpdate = 1;

                Image ecwImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);

                for (int nChunk = 0; nChunk < frames.length; nChunk++) {
                    ChunkFrame f = frames[nChunk];
                    
                    try{
	                    if (bandCount != 3) {
	                        setFileView(file.numBands, bandlist, f);
	                    } else {
	                        setFileView(file.numBands, bandListTriband, f);
	                    }
                    }catch(JNCSInvalidSetViewException exc){
                    	trySupersampling = true;
                    }

                    order = calcMaskAndShift(bandlist, mascara, shl, shr);

                    for (line = 0; line < f.height; line++) {
                        file.readLineRGBA(pRGBArray);
                        pRGBArray = changeBands(order, pRGBArray, mascara, shl, shr);
                        setRGBLine((BufferedImage) ecwImage, f.pos.x, f.pos.y + line, f.width, 1, pRGBArray, 0, f.width);
                    }
                } //Chuncks

                applyAlpha(ecwImage);

                if (frames[0].mustResize && !this.multifile)
                    return resizeImageII(fullSize, ecwImage);
                
                lastRefreshPercent = file.getPercentComplete();

                //System.out.println("Leido al "+lastRefreshPercent+" %.");
                return ecwImage;
            } else { //Caso en el que se actualiza una banda del Image
                EcwFile.nUpdate++;

                for (int nChunk = 0; nChunk < frames.length; nChunk++) {
                    ChunkFrame f = frames[nChunk];

                    if (bandCount != 3) {
                        setFileView(file.numBands, bandlist, f);
                    } else {
                        setFileView(file.numBands, bandListTriband, f);
                    }

                    order = calcMaskAndShift(bandlist, mascara, shl, shr);
                    
                    for (line = 0; line < f.height; line++) {
                        file.readLineRGBA(pRGBArray);
                        pRGBArray = changeBands(order, pRGBArray, mascara, shl, shr);
                        setRGBLine((BufferedImage) img, f.pos.x, f.pos.y + line, f.width, 1, pRGBArray, 0, f.width, origBand, destBandFlag);
                    }
                } //Chuncks

                applyAlpha(img);

                if (frames[0].mustResize && (nUpdate == 3) && this.multifile) {
                    return resizeImageII(fullSize, img);
                }

                lastRefreshPercent = file.getPercentComplete();

                //System.out.println("Leido al "+lastRefreshPercent+" %.");
                return img;
            }
        } catch (com.ermapper.ecw.JNCSException e) { //java.lang.ArrayIndexOutOfBoundsException:
            bErrorOnOpen = true;
            System.err.println("EcwFile JNCS Error en la lnea " + line + "/" +
                               height);
            System.err.println(e.getMessage());
        } catch (java.lang.ArrayIndexOutOfBoundsException e) { //:
            bErrorOnOpen = true;
            System.err.println("EcwFile ArrayIndex Error en la lnea " + line +"/" + height);
            System.err.println(e.getMessage());
        } catch (Exception e) {
            bErrorOnOpen = true;
            errorMessage = e.getMessage();
            System.err.println(errorMessage);
            if(trySupersampling)
            	throw new SupersamplingNotSupportedException();
        }

        return img;
    }

    private Image resizeImageII(Dimension sz, Image image) {
        int w = (int)sz.getWidth();
        int h = (int)sz.getHeight();
    	Image buffImg = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
    	
        //Desplazamiento para la X y la Y leidas. Estas tienen efecto cuando un pixel no empieza a visualizarse
        //justo en su esquina superior izquierda y tiene que ser cortado en la visualizacin.
       
        double currentViewX = (((double) file.width)/(Math.abs(extent.getMax().getX() - extent.getMin().getX())))*(v.minX()-file.originX);
		double currentViewY = (((double) file.height)/(Math.abs(extent.getMax().getY() - extent.getMin().getY())))*(file.originY - v.minY());
        double offsetX = Math.abs(currentViewX - ((int)currentViewX));
        double offsetY = Math.abs(currentViewY - ((int)currentViewY));
        
        int xSrc, ySrc;
        int decr = 2;

        if(v.minX() == extent.minX() || v.maxX() == extent.maxX() || v.minY() == extent.minY() || v.maxY() == extent.maxY()){
        	decr = 0;
        	offsetX = offsetY = 0;
        }

        double scaleW = (double)((double)(image.getWidth(null) - decr) / (double)w);
        double scaleH = (double)((double)(image.getHeight(null) - decr) / (double)h);
        for (int y1 = 0; y1 < h; y1++){
        ySrc = (int) ((y1 * scaleH) + offsetY);
        	for (int x1 = 0; x1 < w; x1++){
        		xSrc = (int) ((x1 * scaleW) + offsetX);
        		try {
        			((BufferedImage) buffImg).setRGB(x1, y1, ((BufferedImage)image).getRGB(xSrc, ySrc));
        		} catch (java.lang.ArrayIndexOutOfBoundsException e) {
        		}
        	}
        }
        
        return buffImg;
    }

    
    private void applyAlpha(Image im) {
        BufferedImage img = (BufferedImage) im;
        int alpha = (getAlpha() & 0xff) << 24;
        int w = img.getWidth();
        int[] line = new int[w];

        for (int j = 0; j < img.getHeight(); j++) {
            img.getRGB(0, j, w, 1, line, 0, w);

            for (int i = 0; i < w; i++)
                line[i] = (alpha | (line[i] & 0x00ffffff));

            img.setRGB(0, j, w, 1, line, 0, w);
        }
    }

    /* (non-Javadoc)
     * @see org.cresques.io.GeoRasterFile#getData(int, int)
     */
    public Object getData(int x, int y, int band) {
        //file.readLineRGBA();
        return null;
    }

    /**
     * Devuelve los datos de una ventana solicitada
     * @param ulX        coordenada X superior izda.
     * @param ulY        coordenada Y superior derecha.
     * @param sizeX        tamao en X de la ventana.
     * @param sizeY tamao en Y de la ventana.
     * @param band        Banda solicitada.
     */
    public byte[] getWindow(int ulX, int ulY, int sizeX, int sizeY, int band) {
        //TODO Nacho: Implementar getWindow de EcwFile
        return null;
    }

    /**
     * Obtiene la zona (Norte / Sur)
     * @return true si la zona es norte y false si es sur
     */
    public boolean getZone() {
        //TODO Nacho: Implementar getZone de EcwFile
        return false;
    }

    /**
     *Devuelve el nmero de zona UTM
     *@return Nmero de zona
     */
    public int getUTM() {
        //		TODO Nacho: Implementar getUTM de EcwFile
        return 0;
    }

    /**
     * Obtiene el sistema de coordenadas geograficas
     * @return Sistema de coordenadas geogrficas
     */
    public String getGeogCS() {
        //TODO Nacho: Implementar getGeogCS de EcwFile
        return new String("");
    }

    /**
     * Devuelve el tamao de bloque
     * @return Tamao de bloque
     */
    public int getBlockSize() {
        //TODO Nacho: Implementar getBlockSize de EcwFile	
        return 1;
    }
    
    /**
	 * Calcula la transformacin que se produce sobre la vista cuando la imagen tiene un fichero .rmf
	 * asociado. En Ecw el origen de coordenadas en Y es el valor mximo y decrece hasta el mnimo.
	 * @param originX Origen de la imagen en la coordenada X
	 * @param originY Origen de la imagen en la coordenada Y
	 */
    public void setExtentTransform(double originX, double originY, double w, double h, double psX, double psY) {
		
	}

    /**
     * Trozo de imagen (Chunk) en que se divide la consulta a la librera,
     * para esquivar el bug#2.
     *
     * @author luisw
     */
    static class ChunkFrame {
        // Ancho mximo (~2500 px)
        final static int MAX_WIDTH = 1536;

        // Alto mximo (no hay lmite)
        final static int MAX_HEIGHT = 1536;
        Point pos;
        Extent v;
        int width;
        int height;
        boolean mustResize = false;

        public ChunkFrame(Extent vista, int w, int h) {
            v = vista;
            width = w;
            height = h;
        }

        /**
         * Calcula el array de chunks (trozos).
         * @param file        Fichero ecw que hay que trocear.
         * @param v Extent total de la vista.
         * @param sz        Tamao total de la vista.
         * @return array de ChunkFrames.
         * @throws JNCSFileNotOpenException
         */
        public static ChunkFrame[] computeFrames(JNCSFile file, Extent v,
                                                 Dimension sz, Extent extent)
                                          throws JNCSFileNotOpenException {
            ChunkFrame[] frames = null;
            ChunkFrame f = null;

            // Calcula el n de chunks (filas y columnas)
            int numCol = (sz.width / MAX_WIDTH);

            // Calcula el n de chunks (filas y columnas)
            int numRow = (sz.height / MAX_HEIGHT);

            if ((sz.width - (numCol * MAX_WIDTH)) > 0) {
                numCol++;
            }

            if ((sz.height - (numRow * MAX_HEIGHT)) > 0) {
                numRow++;
            }

            frames = new ChunkFrame[numCol * numRow];

            JNCSDatasetPoint ptMin = file.convertWorldToDataset(v.minX(), v.minY());
            JNCSDatasetPoint ptMax = file.convertWorldToDataset(v.maxX(), v.maxY());

            //No utilizamos JNCSDatasetPoint porque siempre hace un redondeo por abajo con lo que perdemos precisin. En su lugar
            //utilizamos currentViewM... calculado manualmente y que nos proporciona todos los decimales 
            double currentViewMinX = (((double) file.width)/(Math.abs(extent.getMax().getX() - extent.getMin().getX())))*(v.minX()-file.originX);
            double currentViewMaxX = (((double) file.width)/(Math.abs(extent.getMax().getX() - extent.getMin().getX())))*(v.maxX()-file.originX);
    		double currentViewMinY = (((double) file.height)/(Math.abs(extent.getMax().getY() - extent.getMin().getY())))*(file.originY - v.minY());
    		double currentViewMaxY = (((double) file.height)/(Math.abs(extent.getMax().getY() - extent.getMin().getY())))*(file.originY - v.maxY());
                      
            if ((ptMax.x - ptMin.x) < sz.width) {
                numCol = numRow = 1;
                frames = new ChunkFrame[numCol * numRow];
                int nPixelsX = (int)Math.ceil(Math.abs(currentViewMaxX - currentViewMinX));
                int nPixelsY = (int)Math.ceil(Math.abs(currentViewMaxY - currentViewMinY));
                
                if(v.minX() == extent.minX() || v.maxX() == extent.maxX() || v.minY() == extent.minY() || v.maxY() == extent.maxY()){
                	f = frames[0] = new ChunkFrame(v, nPixelsX, nPixelsY);
                	f.v = new Extent(v);
                }else{
                	f = frames[0] = new ChunkFrame(v, nPixelsX + 1, nPixelsY + 1);
                	double pointEndWcX = v.minX() + (((nPixelsX + 1) * Math.abs(v.maxX() - v.minX())) / nPixelsX);
                	double pointEndWcY = v.maxY() - (((nPixelsY + 1) * Math.abs(v.maxY() - v.minY())) / nPixelsY);
                	f.v = new Extent(v.minX(), v.maxY(), pointEndWcX, pointEndWcY);
                }
                
                f.pos = new Point(0, 0);
                f.mustResize = true;
            } else {
                // Calcula cada chunk
                double stepx = ((double) ptMax.x - ptMin.x) / sz.getWidth();
                double stepy = ((double) ptMax.y - ptMin.y) / sz.getHeight();
                int h = sz.height;

                for (int r = 0; r < numRow; r++) {
                    int w = sz.width;

                    for (int c = 0; c < numCol; c++) {
                        f = new ChunkFrame(null, -1, -1);

                        // Posicin del chunk
                        f.pos = new Point(c * MAX_WIDTH, r * MAX_HEIGHT);

                        // Tamao del chunk
                        f.width = Math.min(MAX_WIDTH, w);
                        f.height = Math.min(MAX_HEIGHT, h);

                        // Extent del chunk
                        int x1 = ptMin.x + (int) (f.pos.x * stepx);
                        int x2 = x1 + (int) (f.width * stepx);
                        int y1 = ptMax.y - (int) (f.pos.y * stepy); 
                        int y2 = y1 - (int) (f.height * stepy); //ptMin.y;
                        JNCSWorldPoint pt1 = file.convertDatasetToWorld(x1, y1);
                        JNCSWorldPoint pt2 = file.convertDatasetToWorld(x2, y2);
                        
                        f.v = new Extent(pt1.x, pt1.y, pt2.x, pt2.y); // Hay que calcularlo
                        frames[(r * numCol) + c] = f;
                        w -= MAX_WIDTH;
                    }

                    h -= MAX_HEIGHT;
                }
            }

            //System.out.println("Hay "+numRow+" filas y "+numCol+" columnas.");
            return frames;
        }
    }
}
