/* Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 *
 * Copyright (C) 2004-5.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 * cresques@gmail.com
 */
package org.cresques.io.raster;

import java.awt.Image;
import java.util.Hashtable;

import org.cresques.geo.ViewPortData;
import org.cresques.px.Extent;


/**
 * Filtro para raster. Ancestro de todos los filtros.
 * @author Luis W. Sevilla (sevilla_lui@gva.es)
 */
public abstract class RasterFilter implements IRasterFilter {
    protected RasterBuf 	raster = null;
    protected RasterStats 	stats = null;
    protected Image 		image = null;
    protected int[] 		px = new int[4];
    protected int 			height = 0;
    protected int 			width = 0;
    protected Hashtable 	params = new Hashtable();
    protected int 			incX = 1;
    protected int 			incY = 1;
    protected ViewPortData	viewPortData = null;
    protected Extent		extent = null;
    
    /**
     * Variable que control la aplicacin o no del filtro. Si est a false aunque est en 
     * la pila el filtro no se ejecutar.
     */
    protected boolean 		exec = true; 

    /**
     * @author Nacho Brodin (brodin_ign@gva.es)
     */
    public static class Kernel{
    	public float[][] kernel = null;
    	public Kernel(float[][] k){
    		this.kernel = k;
    	}
    	
    	public float kernelOperation(Kernel k){
    		float res = 0;
    		for(int i=0;i<3;i++)
    			for(int j=0;j<3;j++)
    				res += kernel[i][j] *  k.kernel[i][j];
    		return res;
    	}
    	
    }

    /**
     * Constructor
     */
    public RasterFilter() {
    }

    /**
     * Aplica el filtro sobre el raster pasado pixel a pixel
     */
    public void execute() {
        pre();
        if (exec) {
            for (int y = 0; y < height; y = y + incY)
                for (int x = 0; x < width; x = x + incX) {
                    process(x, y);
                }
        }
        post();
    }

    /**
     * Aplica el filtro sobre el raster pasado por lineas
     */
    public void executeLines() {
        if (exec) {
            for (int y = 0; y < height; y++)
                processLine(y);
        }
    }

    /**
     * Aade un parmetro al filtro
     * @param name	Clave del parmetro
     * @param param Objeto pasado como parmetro
     */
    public void addParam(String name, Object param) {
        params.put(name, param);
    }

    /**
     * Elimina un parmetro del filtro
     * @param name Clave del parmetro a eliminar
     */
    public void removeParam(String name){
    	params.remove(name);
    }
    
    /**
     * Obtiene un parmetro a partir de la clave
     * @param name Parmetro
     * @return Parmetro
     */
    public Object getParam(String name){
    	return params.get(name);
    }
    
    /**
     * Funcin que contiene el cdigo a ejecutar antes de aplicar el filtro
     */
    abstract public void pre();

    /**
     * Funcin que contiene el cdigo a ejecutar despues de aplicar el filtro
     */
    abstract public void post();

    /**
     * Ejecucin del filtro para un pixel de la imagen
     */
    abstract public void process(int x, int y);

    /**
     * Ejecucin del filtro para una lnea de la imagen
     */
    abstract public void processLine(int y);

    /**
     * Obtiene el tipo de datos del raster de entrada
     */
    abstract public int getInRasterDataType();

    /**
     * Obtiene el tipo de datos del raster de salida
     */
    abstract public int getOutRasterDataType();

    /**
     * Obtiene el resultado del filtro despues de su ejecucin a travs de una clave
     * @param name        clave para obtener un objeto resultado del filtro.
     */
    abstract public Object getResult(String name);

	/**
	 * @param viewPortData The viewPortData to set.
	 */
	public void setViewPortData(ViewPortData viewPortData) {
		this.viewPortData = viewPortData;
	}
	/**
	 * @param extent The extent to set.
	 */
	public void setExtent(Extent extent) {
		this.extent = extent;
	}

	/**
	 * Obtiene true si el filtro va a ser ejecutado o false si no va a serlo
	 * @return
	 */
	public boolean isExec() {
		return exec;
	}

	/**
	 * Asigna el valor a la variable exec. Esta estar a true si el filtro se ejecutar la prxima
	 * vez que se repinte o false si no se ejecuta.
	 * @param exec
	 */
	public void setExec(boolean exec) {
		this.exec = exec;
	}
}
