/*
 * Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 *
 * Copyright (C) 2004-5.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 * cresques@gmail.com
 */
package org.cresques.io.raster;

import java.io.IOException;

	/**
	 * Clase que representa a un conjunto de pixeles con
	 * transparencia. Incluye los rangos de transparencia para cada banda, 
	 * la cadena que va en la lista y la operacin que se realiza entre 
	 * ellos. Un And significa que sern transparentes todos los pixeles
	 * que cumplan con R con G y con B. Un Or significar que tendrn transparencia
	 * todos los pixeles que cumplan con R con G o con B.  
	 * @author Nacho Brodin (brodin_ign@gva.es)
	 *
	 */
	public class TransparencyRange{
		private String 	strEntry = null;
		private int[]	red = null;
		private int[]	green = null;
		private int[]	blue = null;
		private boolean	and = true;
		
		/**
		 * Obtiene la operacin  utilizada
		 * @param and Si es true significa que se usa un AND y false implica
		 * que se usa un OR
		 */
		public boolean isAnd() {
			return and;
		}
		
		/**
		 * Asigna la operacin AND como la utilizada
		 * @param and booleano que si est a true significa que el la operacin
		 * AND es la utilizada como operacin.
		 */
		public void setAnd(boolean and) {
			this.and = and;
		}
		
		/**
		 * Obtiene el intervalo de valores correspondiente a la banda del azul
		 * @return Array bidimensional de enteros correspondiente a la banda del azul
		 */
		public int[] getBlue() {
			return blue;
		}
		
		/**
		 * Asigna el intervalo de valores correspondiente a la banda del azul
		 * @param blue Array bidimensional de enteros correspondiente a la banda del azul
		 */
		public void setBlue(int[] blue) {
			this.blue = blue;
		}
		
		/**
		 * Obtiene el intervalo de valores correspondiente a la banda del verde
		 * @return Array bidimensional de enteros correspondiente a la banda del verde
		 */
		public int[] getGreen() {
			return green;
		}
		
		/**
		 * Asigna el intervalo de valores correspondiente a la banda del verde
		 * @param green Array bidimensional de enteros correspondiente a la banda del verde
		 */
		public void setGreen(int[] green) {
			this.green = green;
		}
		
		/**
		 * Obtiene el intervalo de valores correspondiente a la banda del rojo
		 * @return Array bidimensional de enteros correspondiente a la banda del rojo
		 */
		public int[] getRed() {
			return red;
		}
		
		/**
		 * Asigna el intervalo de valores correspondiente a la banda del rojo
		 * @param red Array bidimensional de enteros correspondiente a la banda del rojo
		 */
		public void setRed(int[] red) {
			this.red = red;
		}
		
		/**
		 * Obtiene la cadena que representa una entrada en la tabla.
		 * @return Cadena que representa una entrada en la tabla
		 */
		public String getStrEntry() {
			return strEntry;
		}
		
		/**
		 * Asigna la cadena que representa una entrada en la tabla.
		 * @param strEntry Cadena que representa una entrada en la tabla.
		 */
		public void setStrEntry(String strEntry) {
			this.strEntry = strEntry;
		}
		
		
		/**
		 * Esta funcin valida la cadena de entrada por medio de una mquina de estados. Valida las
		 * siguientes posibilidades en la cadena de entrada:
		 * <LI>
		 * <UL>(valor_rojo) & (Valor_verde) & (Valor_azul)</UL>
		 * <UL>(valor_rojo) | (Valor_verde) | (Valor_azul)</UL>
		 * </LI>
		 * Despues de la validacin parsea el contenido y carga los parmetros r,g,b con los
		 * intervalos de valores. Estos parmetros deben ser pasados como arrays de enteros de 
		 * dos elementos.  
	     * @param values
	     * @param r	Intervalo de rojo
	     * @param g Intervalo de verde
	     * @param b Intervalo de azul
	     * @return Devuelve true si la operacin usada en los intervalos es un AND y false si es un OR
	     */
		public static boolean stringToInterval(String values, int[] r, int[] g, int[] b) throws IOException {
			int status = 0;
		    int countAnd = 0, countOr = 0;
		    boolean and = true;
		    for(int i=0;i<values.length();i++){
		      char c = values.charAt(i);
		      switch(status){
		      case 0: 	if(c == ' ')status = 0;
		      			else if ((c >= 48 && c <= 57) || c == '*')status = 1;
     					else status = 4;
     					break;
		      case 1: 	if ((c >= 48 && c <= 57) || (c == ' '))status = 1;
     					else if(c == ':') status = 2;
     					else if(c == '&'){
							status = 0;
							countAnd++;
		    	   		}else if(c == '|'){
							status = 0;
							countOr++;
		    	   		}else status = 4;
     					break;
		      case 2:	if (c >= 48 && c <= 57) status = 3;
		      			else status = 4;
		      			break;
		      case 3: 	if ((c >= 48 && c <= 57) || (c == ' '))status = 3;
		      			else if(c == '&'){
		      				status = 0;
		      				countAnd++;
		      			}else if(c == '|'){
		      				status = 0;
		      				countOr++;
		      			}else status = 4;
						break;
		      case 4:	throw new IOException("Error en la cadena de entrada "+status);
		      }
		    }
		    	
		    //Si el analisis es correcto parseamos
		    if((status == 1 || status ==3) && ((countAnd == 2 && countOr == 0)  || 
		    					(countAnd == 0 && countOr == 2)) ){
		    	  String[] s = values.split(" & ");
				  if(s.length < 3){
					  s = values.split(" \\| ");
					  and = false;
				  }
				  ;
				String[] val = s[0].split(":");
				try{
					r[0] = Integer.parseInt(val[0]);
				 	if(val.length == 2)
				 		r[1] = Integer.parseInt(val[1]);
				 	else
						r[1] = Integer.parseInt(val[0]);
				}catch(NumberFormatException e){
					r [0] = -1;r [1] = -1;
				}
				 
				val = s[1].split(":");
				try{
				  	g[0] = Integer.parseInt(val[0]);
					if(val.length == 2)
						g[1] = Integer.parseInt(val[1]);
					else
						g[1] = Integer.parseInt(val[0]);
				}catch(NumberFormatException e){
					g[0] = -1;g[1] = -1;
				}
				  
				val = s[2].split(":");
				try{
					b[0] = Integer.parseInt(val[0]);
					if(val.length == 2)
						b[1] = Integer.parseInt(val[1]);
					else
						b[1] = Integer.parseInt(val[0]);  
				}catch(NumberFormatException e){
					b [0] = -1;b [1] = -1;
				}
		  }else
			 throw new IOException("Error en la cadena de entrada ");
		      
		  return and;
		}
	    
		/**
		 * Muestra los datos del objeto para depuracin
		 *
		 */
		public void show(){
			if(getRed() != null)
				System.out.println(getRed()[0]+" "+getRed()[1]);
			if(getGreen() != null)
				System.out.println(getGreen()[0]+" "+getGreen()[1]);
			if(getBlue() != null)
				System.out.println(getBlue()[0]+" "+getBlue()[1]);
			System.out.println(isAnd());
			System.out.println(getStrEntry());
			System.out.println("--------------------");
		}
	}