package com.iver.cit.gvsig.fmap.edition;

import java.awt.Image;
import java.awt.geom.Rectangle2D;
import java.io.IOException;

import com.hardcode.driverManager.DriverLoadException;
import com.hardcode.gdbms.engine.data.DataSource;
import com.iver.cit.gvsig.fmap.core.IRow;
import com.iver.cit.gvsig.fmap.drivers.DriverIOException;
import com.iver.cit.gvsig.fmap.edition.commands.CommandRecord;
import com.iver.cit.gvsig.fmap.layers.FBitSet;
import com.iver.cit.gvsig.fmap.layers.SelectableDataSource;


/**
 * DOCUMENT ME!
 *
 * @author $author$
 */
public interface IEditableSource {
	/**
     * Mtodo invocado cuando se comienza la edicin, para poner en marcha las
     * estructuras de datos necesarias para la misma, notificar al servidor en
     * protocolos en los que sea necesario, ...
     *
     * @throws EditionException Si no se logra poner la fuente de datos en
     *         edicin
     */
    void startEdition(int sourceType) throws EditionException;

    /**
     * Invocado cuando termina la edicin. En funcin de la clase concreta que
     * implemente este mtodo se generar el fichero con los resultados de la
     * edicin, se realizar una transaccin a la base de datos, etc.
     *
     * @throws EditionException Si no se consiguen llevar a cabo las
     *         modificaciones
     */
   void stopEdition(IWriter writer, int sourceType) throws EditionException;

    /**
     * Cancela la edicin sin escribir los cambios
     *
     * @throws IOException Si se produce un error
     */
    void cancelEdition(int sourceType) throws IOException;

    /**
     * Si el ndice se corresponde a una geometria de las originales de la capa
     * en edicin y no ha sido modificada ni eliminada devuelve la geometria
     * original. Si ha sido modificada debera de buscar en el fichero de
     * expansin y si ha sido eliminada debera devolver null
     *
     * @param index ndice de la geometra.
     *
     * @return Geometra.
     *
     * @throws IOException Si se produce un error con el fichero de expansin
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     */
    IRowEdited getRow(int index) throws DriverIOException, IOException;

    /**
     * Devuelve el nmero de geometrias que hay actualmente en edicin.
     *
     * @return Nmero de geometras.
     *
     * @throws DriverIOException Si se produce un error accediendo a la capa
     * @throws DriverException
     */
    int getRowCount() throws DriverIOException, IOException;

    /**
     * Aade una geometria al fichero de expansin y guarda la correspondencia
     * en una tabla asociada.
     *
     * @param g geometra a guardar.
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     * @throws IOException Si se produce un error con el fichero de expansin
     */
     int addRow(IRow row,String descrip, int sourceType) throws DriverIOException, IOException;

    /**
     * Deshace la ltima accin realizada. Si no hay ms acciones no realiza
     * ninguna accin
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     * @throws IOException Si se produce un error con el fichero de expansin
     */
    void undo() throws DriverIOException, IOException;

    /**
     * Rehace la ltima accin deshecha. Si no hay ms acciones no hace nada
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         filas originales
     */
    void redo() throws DriverIOException, IOException;

    /**
     * Devuelve true si se puede deshacer un comando invocando undo o si por el
     * contrario se ha llegado al principio de la lista de comandos y ya no se
     * pueden deshacer ms comandos
     *
     * @return boolean
     */
    boolean moreUndoCommands();

    /**
     * Devuelve true si se puede rehacer un comando invocando redo o si por el
     * contrario se est al final de la lista de comandos y ya no se pueden
     * deshacer ms comandos
     *
     * @return boolean
     */
    boolean moreRedoCommands();

    /**
     * Elimina una geometria. Si es una geometra original de la capa en
     * edicin se marca como eliminada (haya sido modificada o no). Si es una
     * geometra aadida posteriormente se invalida en el fichero de
     * expansin, para que una futura compactacin termine con ella.
     *
     * @param index ndice de la geometra que se quiere eliminar
     *
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     * @throws IOException Si se produce un error con el fichero de expansin
     */
    void removeRow(int index,String descrip, int sourceType) throws DriverIOException, IOException;

    /**
     * Si se intenta modificar una geometra original de la capa en edicin se
     * aade al fichero de expansin y se registra la posicin en la que se
     * aadi. Si se intenta modificar una geometria que se encuentra en el
     * fichero de expansin (por ser nueva o original pero modificada) se
     * invoca el mtodo modifyGeometry y se actualiza el ndice de la
     * geometria en el fichero.
     *
     * @param index ndice de la geometra que se quiere eliminar
     * @param type TODO
     * @param g Geometra nueva
     * @throws IOException Si se produce un error con el fichero de expansin
     * @throws DriverIOException Si se produce un error accediendo a las
     *         geometras originales
     */
    int modifyRow(int index, IRow row,String descrip, int sourceType)
        throws IOException, DriverIOException;



    /**
     * Compacta el almacenamiento de las geometras que estn en edicin. Tras
     * esta operacin, el orden de las geometras seguramente cambiar y toda
     * llamada a getGeometry devolver una geometra distinta de null, ya que
     * las eliminadas son borradas definitivamente. Hay que tener especial
     * cuidado al invocar este mtodo ya que cualquier tipo de asociacin
     * entre geometras y otro tipo de objetos (comandos de edicin, snapping,
     * ...) que use el ndice de la geometra se ver afectado por ste mtodo
     */
    void compact();

    /**
     * Establece la imagen de las geometras seleccionadas con el fin de que en
     * una edicin interactiva se pueda obtener dicha imagen para simular el
     * copiado, rotado, etc
     *
     * @param i imagen
     */
    // void setSelectionImage(Image i);

    /**
     * Obtiene una imagen con las geometras seleccionadas
     *
     * @return imagen
     */
    // Image getSelectionImage();

    /**
     * DOCUMENT ME!
     */
    void startComplexRow();

    /**
     * DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     * @throws DriverIOException DOCUMENT ME!
     */
    void endComplexRow() throws IOException, DriverIOException;
    public void undoModifyRow(int geometryIndex,int previousExpansionFileIndex, int sourceType) throws IOException, DriverIOException;
    public IRow doRemoveRow(int index, int sourceType)throws DriverIOException, IOException;
    public int doModifyRow(int index, IRow feat, int sourceType)throws IOException, DriverIOException;
    public int doAddRow(IRow feat, int sourceType) throws DriverIOException, IOException;
    public void undoRemoveRow(int index, int sourceType)throws IOException, DriverIOException;
    public void undoAddRow(int index, int sourceType)throws DriverIOException, IOException;
    public SelectableDataSource getRecordset() throws DriverLoadException;
    public boolean isEditing();
    public FBitSet getSelection();
    public CommandRecord getCommandRecord();

	public void addEditionListener(IEditionListener listener);

	public void removeEditionListener(IEditionListener listener);

}
