package com.hardcode.gdbms.engine.data;

import com.hardcode.gdbms.engine.data.driver.DriverException;
import com.hardcode.gdbms.engine.data.driver.ReadAccess;
import com.hardcode.gdbms.engine.data.edition.DataWare;
import com.hardcode.gdbms.engine.data.persistence.Memento;
import com.hardcode.gdbms.engine.data.persistence.MementoException;
import com.hardcode.gdbms.engine.values.Value;
import com.hardcode.gdbms.engine.values.ValueCollection;

import java.io.IOException;


/**
 * Interfaz que define los origenes de datos para operaciones internas del
 * motor de base de datos
 *
 * @author Fernando Gonzlez Corts
 */
public interface DataSource extends ReadAccess {
    /**
     * Avisa al driver de que van a realizarse operaciones sobre l. El driver
     * deber de abrir el fichero, conectar a la base de datos
     *
     * @throws DriverException Si se produce algn error abriendo el DataSource
     */
    void start() throws DriverException;

    /**
     * Cierra el DataSource. Todo acceso que no se site temporalmente entre un
     * start y un stop tendr un comportamiento indeterminado
     *
     * @throws DriverException Si se produce un error cerrando la fuente de
     *         datos
     */
    void stop() throws DriverException;

    /**
     * Devuelve el nombre del DataSource
     *
     * @return nombre de la tabla
     */
    String getName();

    /**
     * Devuelve el filtro que result de la clusula where de la instruccin
     * que di como resultado este DataSource.
     *
     * @return Filtro de la clusula where o null si el DataSource no es
     *         resultado de una instruccin con clusula where
     *
     * @throws IOException Si se produce un error accediendo a las estructuras
     *         de datos internas
     *
     * @deprecated
     */
    long[] getWhereFilter() throws IOException;

    /**
     * gets a reference to the factory object that created the DataSource
     *
     * @return DataSourceFactory
     */
    DataSourceFactory getDataSourceFactory();

    /**
     * Gets a memento object with the current status of the DataSource
     *
     * @return DataSourceMemento
     *
     * @throws MementoException If the state cannot be obtained
     */
    Memento getMemento() throws MementoException;

    /**
     * Sets the DataSourceFactory that created the DataSource instance
     *
     * @param dsf DataSourceFactory
     */
    public void setDataSourceFactory(DataSourceFactory dsf);

    /**
     * Sets the source information of the DataSource
     *
     * @param sourceInfo instance with the info
     */
    public void setSourceInfo(SourceInfo sourceInfo);

    /**
     * Gets the source information of the DataSource
     *
     * @return instance with the info
     */
    public SourceInfo getSourceInfo();

    /**
     * Gets the string representation of this DataSource
     *
     * @return String
     *
     * @throws DriverException
     */
    public String getAsString() throws DriverException;

    /**
     * Removes from the system the data source this DataSource instance
     * represents. No method can be called and no DataSource instance can be
     * obtained from the system after calling this method.
     *
     * @throws DriverException if the DataSource original system could not be
     *         cleaned properly
     */
    public void remove() throws DriverException;

    /**
     * Gets the field ids of the DataSource primary keys fields
     *
     * @return int[] or null if the DataSource is not editable
     *
     * @throws DriverException If the operation fails
     */
    public int[] getPrimaryKeys() throws DriverException;

    /**
     * Gets the value of the primary key
     *
     * @param rowIndex row
     *
     * @return ArrayValue with the values of the primary key fields
     *
     * @throws DriverException Si se produce un error accediendo al DataSource
     */
    public abstract ValueCollection getPKValue(long rowIndex)
        throws DriverException;

    /**
     * Gets the name of the fieldIdth primary key field
     *
     * @param fieldId ndice del campo cuyo nombre se quiere obtener
     *
     * @return String
     *
     * @throws DriverException Si se produce algn error accediendo al
     *         DataSource
     */
    public abstract String getPKName(int fieldId) throws DriverException;

    /**
     * Get's the names of the primary key fields
     *
     * @return String[]
     *
     * @throws DriverException If the access fails
     */
    public abstract String[] getPKNames() throws DriverException;

    /**
     * Returns the ith field type. Must be a java.sql.Types constant
     *
     * @param i ndice del campo cuyo tipo se quiere conocer
     *
     * @return Class
     *
     * @throws DriverException Si se produce un error accediendo a los tipos de
     *         datos
     */
    public abstract int getPKType(int i) throws DriverException;

    /**
     * Return the number of fields that are primary key of this DataSource
     *
     * @return
     *
     * @throws DriverException
     */
    int getPKCardinality() throws DriverException;

    /**
     * Gets the value of all fields at the specified row
     *
     * @param rowIndex index of the row to be retrieved
     *
     * @return Value[]
     *
     * @throws DriverException If the access fails
     */
    Value[] getRow(long rowIndex) throws DriverException;

    /**
     * Gets the field names array
     *
     * @return String[]
     *
     * @throws DriverException if the access fails
     */
    String[] getFieldNames() throws DriverException;

    /**
     * Obtiene el indice de un campo a partir de su nombre o -1 si no existe un
     * campo con ese nombre
     *
     * @param fieldName Nombre del campo
     *
     * @return Indice del campo con el nombre dado o -1 si el campo no existe
     *
     * @throws DriverException Si se produce un error accediendo a los datos
     */
    int getFieldIndexByName(String fieldName) throws DriverException;

    /**
     * Gets a DataWare over the same data source of this DataSource
     *
     * @param mode One of DataSourceFactory.DATA_WARE_DIRECT_MODE,
     * DataSourceFactory.DATA_WARE_COHERENT_ROW_ORDER
     *
     * @return DataWare
     */
    DataWare getDataWare(int mode) throws DriverException;
    
    /**
     * 
     */
    boolean isVirtualField(int fieldId) throws DriverException;
}
